# init.s: The program entry point and trap handler.

# Global procedures and functions defined in this file:
.globl init
.globl trap_interrupt

# Global procedures and functions called in this file:
.globl transmit_string
.globl init_input_buffer
.globl init_interrupts
.globl init_serial_port
.globl init_display
.globl init_pio
.globl init_servo
.globl init_calibration
.globl init_driver
.globl main
.globl transmit_hex_byte

# Constants used in this file

itc_register = 0x34
initial_stack_value = 0xffff



.text
init:           di # disable interrupts during initialisation.

                # Check for a TRAP interrupt.
                in0         a,(itc_register)
                bit         7,a
                jr          nz,trap_interrupt

                # No TRAP. Continue.
                ld          sp,initial_stack_value

                call        init_input_buffer
                call        init_pio
                call        init_interrupts
                call        init_serial_port
                call        init_display
				call		init_servo
				call		init_calibration
				call		init_driver

                # Go to main program execution point
                call        main

                # Reset the machine
                rst         0

# This block of code handles traps. A TRAP interrupt is generated
# when the machine reads an invalid opcode. The code first prints
# the value of the program counter at the point where the
# problem occurred, and then prints out the stack at that time.
# This helps debug problems with the program.
#
# A user can CALL trap_interrupt directly in order to simulate
# a trap. This is be useful in a few situations, e.g. causing
# a trap on "Divide by Zero" or a similar error.

trap_interrupt:
                # A TRAP interrupt occurred. Do a stack dump
                # to aid debugging.

                pop         bc      # Load PC at trap location

                # The actual trap location is the loaded PC
                # value minus one if the ITC register UFO flag
                # (bit 6) is zero. It is the loaded PC minus two
                # if the ITC register UFO flag is one.
                dec         bc

                bit         6,a
                jr          z,pc_is_correct

                dec         bc
pc_is_correct:  
                # Print out the trap notification string
                ld          de,trap_string
                push        de
                call        transmit_string
                pop         de

                # Print out the trap location.. high byte first
                push        bc
                call        transmit_hex_byte
                pop         bc

                # and the low byte
                ld          b,c
                push        bc
                call        transmit_hex_byte
                pop         bc

                # Now begin a stack dump
                # The stack dump stops when SP reaches the
                # initial stack value.

                # Check the current SP value.
                # Does SP equal it's initial value?
stack_dump:     ld          hl,initial_stack_value
                or          a           # clear carry flag before SBC
                sbc         hl,sp

                # ... ie. is hl zero?
                ld          a,h
                or          l
                cp          0
                jr          z,done_dumping

                # No, there's still more stack to dump
                pop         bc

                # Print out the new stack value string
                ld          de,stack_value_string
                push        de
                call        transmit_string
                pop         de

                # Print out the stack value's high byte
                push        bc
                call        transmit_hex_byte
                pop         bc

                # and it's low byte
                ld          b,c
                push        bc
                call        transmit_hex_byte
                pop         bc

                # carry on dumping.
                jr          stack_dump


done_dumping:   # Stack dump complete. Just halt.
                halt


trap_string:    "\r\nA TRAP Interrupt occurred at location: 0x\0"

stack_value_string: 
                "\r\n0x\0"



