# ir.s: Infra-red input code. The procedure in this file attempts
# to identify which IR beacon, if any, the detector is facing.
#
# The procedures and functions in this file should not be called
# from interrupts (with the exception of ir_signal_interrupt which
# is itself an ISR).

# Global procedures and functions defined in this file:
.globl ir_listen
.globl ir_signal_interrupt
.globl ir_listen_in_background 
.globl ir_signal_was_received

# Global procedures and functions called in this file:
.globl lightweight_pio_sample_input
.globl pio_reset_ir_flipflops
.globl sleep
.globl sleep_again
.globl enable_int0
.globl disable_int0

.globl transmit_character
.globl transmit_hex_byte


# Constants used in this file:
num_samples = 80 # 70
max_wait_samples = 80
sample_time = 0x6c			# Found to be the best sample 
							# time to use by experiment.


.bss
# Global variables used in this file:

# The ir_receive_buffer stores temporary data received by ir_listen.
# It is safe to store this data in a global variable because hardware
# constraints prevent two copies of ir_listen running simultaneously
# anyway - ir_listen cannot be called from an interrupt.

ir_receive_buffer:				.space num_samples + 1

# This flag is set to true by the ir_signal_interrupt ISR.
ir_signal_was_received_flag:	.space 1

.text

# ir_listen: Listen on an IR port for an incoming signal and decode any
# signal received. This function returns 0 if no signal was received
# or the signal could not be decoded correctly. Otherwise, it returns
# a number from 1 to 8 to represent the beacon the detector is facing.
# The function is guaranteed to return within about 50 milliseconds,
# depending on the value of num_samples and max_wait_samples, since
# it will time out if correct data is not received.
#
# The function should be called with the number of the port to listen on.
# This should be 1 or 2 depending on which detector should be used.
#
# Usage:
#       ld a,<ir port number - 1 or 2>
#       push af
#       call ir_listen
#       pop af
#       <A contains the incoming signal number or zero>
ir_listen:
                push        ix
                ld          ix,0x0000
                add         ix,sp

				push		af
				push		bc
				push		de
				push		iy

                # Load the number of the IR port into D.
				ld			d,(ix+5)

				# Zero the return value by default
				ld			a,0
				ld			(ix+5),a

                # Wait for a single sample. The main point of
                # this is to program the sleep timer, so that
                # we can use the sleep_again procedure next time
                # to repeat the same sleep.
                ld          bc,sample_time
				push		bc
				call		sleep
                pop         bc

                # IY is the buffer pointer. It points to the
                # current position in the IR receive buffer.
				ld			iy,ir_receive_buffer

                # Start sampling, until either some data is detected,
                # or a timeout of "max_wait_samples" is reached.
				ld			b,max_wait_samples

loop_while_nothing_received:			
                # Sample the input
				call		lightweight_pio_sample_input
                and         d       # filter out the IR port we're interested in

                # If a signal is being received, then jump to the
				# signal handling loop
				jr			nz,we_get_signal

                # Otherwise, wait for 'sample_time' and sample again
				call		sleep_again
				djnz		loop_while_nothing_received

                # The "max_wait_samples" timeout has expired.
                # Nothing was received, so leave the procedure.
				jr			nothing_received

we_get_signal:  # A signal is being received.
                ld          c,0     		# C stores the current sample number

receiving_mark_bits:
                inc         c				# Increase sample number
				ld			a,c
				cp			num_samples		# Have we got num_samples samples yet?
				jr			z,done_sampling	# If so, stop sampling.

                call        sleep_again

				call		lightweight_pio_sample_input
                and         d       # filter out the IR port we're interested in

				# The zero flag is set if the bit received was a space.
                jr          nz,receiving_mark_bits
			
				# A space was received. Record the sample number of the
				# last mark bit and move to the space loop.

                ld          (iy),c
                inc         iy

receiving_space_bits:
                inc         c				# Increase sample number
				ld			a,c
				cp			num_samples		# Have we got num_samples samples yet?
				jr			z,done_sampling	# If so, stop sampling.

                call        sleep_again

				call		lightweight_pio_sample_input
                and         d       # filter out the IR port we're interested in

				# The zero flag is clear if the bit received was a mark.
                jr          z,receiving_space_bits

				# A mark was received. Record the sample number of the
				# last space bit, and move back to the mark loop.

                ld          (iy),c
                inc         iy

                jr			receiving_mark_bits

done_sampling:	# We have received the appropriate number of samples.
				# Mark the end of the sample buffer with a zero.
				ld			(iy),0

				# Attempt to decode the buffer.
				push		af
				call		ir_decode_buffer
				pop			af

				# Return A
				ld			(ix+5),a
                
nothing_received:
				pop			iy
                pop         de
				pop			bc
				pop			af
				pop			ix
				ret


# ir_decode_buffer: Decode some signal that has been received into the 
# ir_receive_buffer. Output a beacon number (1-8) if one was detected,
# 0 otherwise. 
#
# Usage:
#		push af
#       call ir_decode_buffer
#       pop af
#       <result is in A>

ir_decode_buffer:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                push        bc
                push        de
				push		iy

				# IY points to the current place in the buffer.
				ld			iy,ir_receive_buffer

				# Convert the mark and space timings in IY to a
				# bit field. Each timing is divided by 2 since
				# a signal of length 2 or 3 almost always equates
				# to a single mark or space bit, and a signal
				# of length 4 or 5 almost always equates to 2 mark
				# or space bits, and so on...

				# Each number in IY is relative to the
				# previous number so keep the previous number at
				# each stage.

				# C stores the output bit field
				ld			c,0
				# D stores previous number from IY
				ld			d,0
				# E is a flag set to TRUE for a Mark and FALSE for a Space.
				# The initial bit in the buffer is always a mark, so E is true
				# to begin with.
				ld			e,0xff

decoder_loop:	ld			a,(iy)
				cp			0					# if (IY) = 0, then the buffer end
				jr			z,run_out_of_data	# has been reached

				ld			b,a					# (IY) stored in A and B.
				sub			d					# Subtract previous number from this
												# number
				ld			d,b					# Update previous number register

				# A contains the number of samples in this mark or space.
				# Divide this by 2 to get the actual number of bits
				srl			a

				# A must be > 0 so ensure that it is at least 1.
				cp			0
				jr			nz,number_of_bits_is_nonzero

				ld			a,1

number_of_bits_is_nonzero:
				# Add the required number of bits to the output bit field.
				# We shift the bits in at the most significant end 
				# (left hand side).

				ld			b,a		# B bits to be added
add_bits_to_bit_field:
				sla			c		# shift bit field left

				# Should the new bit be one or zero? Get it from E,
				# which is the Mark/Space flag.
				ld			a,e
				and			0x01	# Keep only the 1st bit of E.
				or			c		# Put that bit into C.
				ld			c,a

				# Keep adding bits while B > 0.
				djnz		add_bits_to_bit_field

				# Now, does C contain a valid IR signal? 
				# valid signal:
				#   0  1  0  n  n  n  p  1
				#  8th ................ 1st bit
				#         BEACON# PARITY
                bit         7,a             # unused bit must be 0.
                jr          nz,not_valid_signal
                bit         6,a             # start bit must be 1.
                jr          z,not_valid_signal
                bit         5,a             # zero
                jr          nz,not_valid_signal
                bit         0,a             # stop bit must be 1.
                jr          z,not_valid_signal

				# Ok, all the fixed bits are correct. Check the parity
				# bit (bit 1). Parity is set so as to make the total
				# of bits 1 through 6 even.
				# Mask bits 1 through 6:
				and			0x7e

				# And then check that the even parity bit is set in
				# the flags word:
				jp			po,not_valid_signal		# (jump if parity is odd)

				# Valid signal received!
				# Marshal the beacon number into the appropriate place.
				# It currently occupies bits 2, 3, and 4.
				# So shift right twice.
				srl			a
				srl			a
				and			0x07		# Remove all other bits
				inc			a			# Add 1. Now A can be returned.
                jr          valid_signal_received

not_valid_signal:
				# Move to next place in the buffer
				inc			iy					
				
				# Invert the Mark/Space flag E
				ld			a,e
				cpl
				ld			e,a

				jr			decoder_loop

run_out_of_data:
				ld			a,0		# No valid signal received.

valid_signal_received:
				# The signal received, or zero, is in A

				ld			(ix+5),a

				# We're done.
				
				pop			iy
                pop			de
                pop			bc
                pop			af
                pop			ix
				ret


# ir_signal_interrupt: This ISR is called when an IR signal is received.
# This is not used in the partially hardware implementation of the IR
# decoder.
ir_signal_interrupt:
                di
				call		disable_int0
                ei
                reti

# ir_listen_in_background: This procedure instructs the IR receiver
# to listen for a signal in the background while other operations proceed.
# This is done by allowing IR signals to generate an interrupt.
# The caller can check if an IR signal was received by calling
# ir_signal_was_received.
ir_listen_in_background: 
                call        pio_reset_ir_flipflops
                ret

# ir_signal_was_received: This function returns non-zero if an IR signal was
# received since the last call to ir_listen_in_background. Otherwise,
# it returns zero. The number returned is actually a bitfield. If the
# 1st bit is set, then the signal was received from the first IR 
# detector. If the 2nd bit is set, then the signal was received from the
# second, etc.
# Usage:
#		push af
#       call ir_signal_was_received
#       pop af
#       <A is non-zero if IR signal was received>
ir_signal_was_received:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                call        lightweight_pio_sample_input
                and         0xf0
                srl         a
                srl         a
                srl         a
                srl         a
				ld			(ix+5),a
                pop         af
                pop         ix
                ret
