# output.s: Formatted output procedures.
# The procedures in this file enable the user to send strings
# and numbers to the display or serial line.
# - The transmit* procedures send data down the serial line.
# - The display* procedures write data on the display.


# Global procedures and functions defined in this file:
.globl transmit_string
.globl transmit_hex_byte
.globl transmit_newline
.globl display_string
.globl display_hex_byte

# Global procedures and functions called in this file:
.globl transmit_character
.globl display_character


# lightweight_general_string_handler: This abstract lightweight
# subroutine calls a given procedure once for each character in a
# zero terminated string.
# AF, BC, DE, and HL may all be destroyed by the use of this
# lightweight subroutine.
# Usage:        Load pointer to start of string into BC
#               Load pointer to the character handling procedure into HL.
#               Call lightweight_general_string_handler
#
lightweight_general_string_handler:
                ld          de,lightweight_general_string_handler_return_point
            
                # Read characters from the string
                # until a zero is read.
lightweight_general_string_handler_loop:
                # Read a character...
                ld          a,(bc)
                
                # is accumulator zero?
                cp          0
                jr          z,lightweight_general_string_handler_loop_end

                push        af
                # Call the character handling procedure.
                # We can't use CALL unfortunately because
                # that can only call absolute addresses.
                # But.. we can simulate CALL by PUSHing a reasonable
                # PC value onto the stack (from DE), and then jumping to
                # the appropriate location.
                push        de
                jp          (hl)

lightweight_general_string_handler_return_point:
                pop         af

                # next character..
                inc         bc
                jr          lightweight_general_string_handler_loop

lightweight_general_string_handler_loop_end:
                # Finished...
                ret



# lightweight_general_hex_handler: This abstract procedure calls a procedure
# passed to it in order to output the hex digits provided by the
# caller. It is designed to be a flexible way to output any number
# of hexadecimal digits to any output device.
# Usage:
#               push        <character_handling_procedure>
#               push        <number of hex digits to output>
#               push        <start address of hex digits>
#               call        general_hex_handler

# lightweight_general_hex_handler: This abstract lightweight
# subroutine calls a given procedure once for DE nibbles from
# a starting address BC.
# AF, BC, DE, and HL may all be destroyed by the use of this
# lightweight subroutine.
# Usage:        Load pointer to start of digits into BC
#               Load pointer to the character handling procedure into HL.
#               Load the number of digits to output into DE.
#               Call lightweight_general_hex_handler
#
lightweight_general_hex_handler:

lightweight_general_hex_handler_loop:
                # DE keeps a count of the number of hex digits
                # left to output. If DE is even, then we output
                # the high nibble of (BC). If DE is odd, then
                # the low nibble is output instead. Check the 1st
                # bit of E to see if DE is even or odd.

                ld          a,(bc)
                bit         0,e
                jr          nz,lightweight_general_hex_handler_odd

                # move high nibble of A to the low nibble.
                rr          a
                rr          a
                rr          a
                rr          a

                # If DE is odd, then increment BC to read the
                # next digit. 

                dec         bc

lightweight_general_hex_handler_odd:
                inc         bc      # BC is incremented only if DE is odd.

                # remove all unused bits
                and         0x0f

                # is A less than 10?
                # if so, display it as a number: 0-9
                cp          10
                jp          m,lightweight_general_hex_handler_number

                # if not, display it as a letter.
                # 10 becomes 'a' which is character 97 in ASCII
                add         a,97 - 10

                jr          lightweight_general_hex_handler_finish

lightweight_general_hex_handler_number:
                # decode it as a number.
                # 0 becomes '0' which is character 48 in ASCII
                add         a,48

lightweight_general_hex_handler_finish:
                push        de # save the contents of DE

                ld          de,lightweight_general_hex_handler_return_point

                # send that character
                push        af
                # Call the character handling procedure.
                # See the above comments in the 
                # lightweight_general_string_handler procedure.
                push        de
                jp          (hl)

lightweight_general_hex_handler_return_point:
                pop         af

                pop         de      # restore DE

                dec         de      # move to next character..

                # is DE zero?
                ld          a,d
                or          e
                cp          0
                jr          nz,lightweight_general_hex_handler_loop

                # Yes, DE is now zero. We are done.

                ret





# transmit_string: Serial communications procedure.
# Sends a null-terminated string of any length to the serial port.
# A pointer to the start of the string must be pushed onto
# the stack before this procedure is called. 
transmit_string:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                push        bc
                push        de
                push        hl

                # Select the procedure that will be
                # used to transmit the string.
                ld          hl,transmit_character

                # Load pointer to start of the string
                ld          c,(ix+4)
                ld          b,(ix+5)
				
                # Do transmission
                call        lightweight_general_string_handler
                
                # Done..
                pop         hl
                pop         de
                pop         bc
                pop         af
                pop         ix
                ret

# display_string: Display output procedure
# Writes a null-terminated string of any length to the display.
# A pointer to the start of the string must be pushed onto
# the stack before this procedure is called. 
display_string:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                push        bc
                push        de
                push        hl

                # Select the procedure that will be
                # used to transmit the string.
                ld          hl,display_character

                # Load pointer to start of the string
                ld          c,(ix+4)
                ld          b,(ix+5)

                # Do transmission
                call        lightweight_general_string_handler

                # Done..
                pop         hl
                pop         de
                pop         bc
                pop         af
                pop         ix
                ret

# transmit_hex_byte: Serial communications procedure.
# Sends an 8 bit hex number to serial port 1.
# The serial port must have been initialised first. The number
# is passed by stack, e.g.
#               ld          a,<number>
#               push        af
#               call        transmit_hex_byte
#               pop         af

transmit_hex_byte:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                push        bc
                push        de
                push        hl

                # Load BC with pointer to start of the hex string: which is ix+5
                push        ix          # Move IX to HL
                pop         hl
                ld          bc,5        # Add 5 to HL
                add         hl,bc
                push        hl          # Move HL to BC
                pop         bc

                # Select the procedure that will be
                # used to transmit the string.
                ld          hl,transmit_character


                # output only 2 hex digits.
                ld          de,2

                call        lightweight_general_hex_handler

                # Done..
                pop         hl
                pop         de
                pop         bc
                pop         af
                pop         ix
                ret


# display_hex_byte: Draws an 8 bit hex number on the display.
# The number is passed by stack, e.g.
#               ld          a,<number>
#               push        af
#               call        display_hex_byte
#               pop         af

display_hex_byte:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                push        bc
                push        de
                push        hl

                # Load BC with pointer to start of the hex string: which is ix+5
                push        ix          # Move IX to HL
                pop         hl
                ld          bc,5        # Add 5 to HL
                add         hl,bc
                push        hl          # Move HL to BC
                pop         bc

                # Select the procedure that will be
                # used to transmit the string.
                ld          hl,display_character

                # output only 2 hex digits.
                ld          de,2

                call        lightweight_general_hex_handler

                # Done..
                pop         hl
                pop         de
                pop         bc
                pop         af
                pop         ix
                ret

newline_string: "\r\n\0"

# transmit_newline: Print a newline to the serial port. There is no
# equivalent display_newline procedure - to move to the next line on
# the display, use display_move_to.
# This procedure takes no parameters.
transmit_newline:
                push        bc
                ld          bc,newline_string
                push        bc
                call        transmit_string
                pop         bc
                pop         bc
                ret

