# pio.s: Low-level routines to control the 82C55A PIO IC.

# Global procedures and functions defined in this file:
.globl pio_set_output
.globl lightweight_pio_sample_input
.globl pio_reset_ir_flipflops
.globl init_pio

# Global procedures and functions called in this file:

# Constants used in this file:

pio_port_a = 0xb0		# Port A: used for servo output
pio_port_b = 0xb1       # Port B: used for IR flip flop control
pio_port_c = 0xb2		# Port C: used for infra-red input
pio_control_port = 0xb3


# pio_set_output: Change the output on port A of the PIO.
# This is typically used to control the servo (but may also
# be useful for debugging).
pio_set_output:
                push        ix
                ld          ix,0x0000
                add         ix,sp

                push        af
                ld          a,(ix+5)
                out0        (pio_port_a),a
                pop         af
                
                pop         ix
                ret

# pio_reset_ir_flipflops: Send a signal to PIO port B to reset
# the flipflops used by the IR decoder (see hardware diagram).
pio_reset_ir_flipflops:
                push        af
                ld          a,0xff
                out0        (pio_port_b),a
                inc         a
                out0        (pio_port_b),a
                pop         af
                ret

# lightweight_pio_sample_input: Sample the input from both
# infra-red lines and store the result in A. This lightweight
# subroutine destroys the A and F registers.
#
# The 1st bit of A is set if the first IR line is active.
# The 2nd bit of A is set if the second IR line is active.
# The 5th bit of A is set if the first IR flip flop has been activated.
# The 6th bit of A is set if the second IR flip flop has been activated.
# All the other bits are undefined, so the caller should logically
# AND the returned value to find which IR lines were active.
#
lightweight_pio_sample_input:
                in0         a,(pio_port_c)
                cpl         # invert every bit of A.
                ret

init_pio:
                push        af

                # control word.
                # (8th bit one) Mode Set Flag is set.
                # (6th, 7th bits zero) Group A in Mode 0
                # (5th bit zero) Port A: output
                # (4th bit one) Port C (upper): input
                # (3rd bit zero) Group B in Mode 0
                # (2nd bit zero) Port B: output
                # (1st bit one) Port C (lower): input
                # Total: 10001001: 0x89

                ld          a,0x89
                out0        (pio_control_port),a

                
				# Output zero to start with
                ld          a,0
                push        af
                call        pio_set_output
                pop         af

                call        pio_reset_ir_flipflops

                pop         af
                ret




